from __future__ import annotations

import email.utils
import mimetypes
import socket
import typing as t
from email.message import EmailMessage


class Attachment:
    """Structured representation of an email attachment.

    :param data: Text or bytes data to attach.
    :param filename: Filename to show for the attachment.
    :param mimetype: Mimetype describing the attached data. Defaults to guessing
        from ``filename`` if possible, or ``text/plain`` for text data or
        ``application/octet-stream`` for bytes data.
    """

    def __init__(
        self,
        data: str | bytes,
        *,
        filename: str | None = None,
        mimetype: str | None = None,
    ):
        if mimetype is None and filename is not None:
            guess = mimetypes.guess_type(filename)[0]

            if guess is not None:
                mimetype = guess

        if mimetype is None:
            if isinstance(data, str):
                mimetype = "text/plain"
            else:
                mimetype = "application/octet-stream"

        self.data = data
        """Text or bytes data to attach."""

        self.filename = filename
        """Filename to show for the attachment."""

        self.mimetype: str = mimetype
        """Mimetype describing the attached data. Defaults to guessing from
        ``filename`` if possible, or ``text/plain`` for text data or
        ``application/octet-stream`` for bytes data.
        """

        self._cid: str | None = None

    @property
    def cid(self) -> str:
        """The id used to refer to the inline attachment in HTML. Generated on
        access if not set.
        """
        if self._cid is None:
            self._cid = email.utils.make_msgid(domain=local_hostname())

        return self._cid

    @cid.setter
    def cid(self, value: str | None) -> None:
        self._cid = value

    def add_to_mime(self, message: EmailMessage, *, inline: bool = False) -> None:
        """Add this attachment to the given :class:`email.message.EmailMessage.
        When attaching inline, include the :attr:`cid`, otherwise the
        :attr:`filename`.

        :param message: The message to attach to.
        :param inline: Attach inline for linking from HTML.
        """
        kwargs: dict[str, t.Any] = {}
        main_type, _, kwargs["subtype"] = self.mimetype.partition("/")
        data: str | bytes = self.data

        if isinstance(data, str):
            if main_type != "text":
                # For example application/json from a .json filename.
                # message.add_attachment insists on text/* mimetypes for str
                # data, so encode it as utf8 to get around that.
                data = data.encode()
                kwargs["maintype"] = main_type
        else:
            kwargs["maintype"] = main_type

        if inline:
            kwargs["cid"] = self.cid
            message.add_related(data, **kwargs)
        else:
            if self.filename:
                kwargs["filename"] = self.filename

            message.add_attachment(data, **kwargs)


_local_hostname: str | None = None
"""Cached value for :func:`local_hostname`."""


def local_hostname() -> str:
    """Return the result of :func:`socket.getfqdn`, caching the value for
    subsequent calls. Used to generate attachment cids and with SMTP.
    """
    global _local_hostname

    if _local_hostname is None:
        _local_hostname = socket.getfqdn()

    return _local_hostname



from __future__ import annotations

import asyncio
from smtplib import SMTP
from unittest.mock import create_autospec
from unittest.mock import MagicMock
from unittest.mock import patch

from email_simplified import Message
from email_simplified import SMTPEmailHandler


def test_tls_port() -> None:
    assert SMTPEmailHandler(port=465).use_tls


def test_tls_disables_starttls() -> None:
    assert not SMTPEmailHandler(use_tls=True, use_starttls=True).use_starttls


def test_tls_context() -> None:
    assert SMTPEmailHandler(use_tls=True).tls_context
    assert SMTPEmailHandler(use_starttls=True).tls_context


def test_from_config() -> None:
    assert SMTPEmailHandler.from_config({"port": 1025, "invalid": True}).port == 1025


@patch("email_simplified.handlers.smtp.SMTP", autospec=True)
def test_connect(smtp_cls: MagicMock) -> None:
    handler = SMTPEmailHandler()
    client: MagicMock

    with handler.connect() as client:  # type: ignore[assignment]
        pass

    assert smtp_cls.call_args.kwargs.keys() == {
        "host",
        "port",
        "local_hostname",
        "timeout",
    }
    client.starttls.assert_not_called()
    client.login.assert_not_called()


@patch("email_simplified.handlers.smtp.SMTP", autospec=True)
def test_connect_setup(smtp_cls: MagicMock) -> None:
    handler = SMTPEmailHandler(use_starttls=True, username="a", password="b")
    client: MagicMock

    with handler.connect() as client:  # type: ignore[assignment]
        pass

    client.starttls.assert_called()
    client.login.assert_called()


@patch("email_simplified.handlers.smtp.SMTP_SSL", autospec=True)
def test_connect_tls(smtp_ssl_cls: MagicMock) -> None:
    handler = SMTPEmailHandler(use_tls=True)

    with handler.connect():
        pass

    assert "context" in smtp_ssl_cls.call_args.kwargs


@patch.object(SMTPEmailHandler, "connect")
def test_send(connect: MagicMock) -> None:
    ctx = create_autospec(SMTP, instance=True)
    connect.return_value.__enter__.return_value = ctx
    handler = SMTPEmailHandler()
    handler.send(
        [
            Message(subject="a"),
            Message(subject="b", from_addr="a@example.test").to_mime(),
        ]
    )
    connect.assert_called()
    assert ctx.send_message.call_count == 2


@patch.object(SMTPEmailHandler, "connect")
def test_send_empty(connect: MagicMock) -> None:
    handler = SMTPEmailHandler()
    handler.send([])
    connect.assert_not_called()


@patch.object(SMTPEmailHandler, "connect")
def test_send_batch(connect: MagicMock) -> None:
    ctx = create_autospec(SMTP, instance=True)
    connect.return_value.__enter__.return_value = ctx
    handler = SMTPEmailHandler(recipients_per_message=4)
    handler.send([Message(subject="a", to=[f"a{x}@example.test" for x in range(10)])])
    assert ctx.send_message.call_count == 3


@patch.object(SMTPEmailHandler, "connect")
def test_send_async(connect: MagicMock) -> None:
    ctx = create_autospec(SMTP, instance=True)
    connect.return_value.__enter__.return_value = ctx
    handler = SMTPEmailHandler()
    asyncio.run(handler.send_async([Message(subject="a")]))
    connect.assert_called()
    assert ctx.send_message.call_count == 1



from __future__ import annotations

import collections.abc as cabc
import html.parser
import typing as t
from email.headerregistry import Address
from email.headerregistry import AddressHeader
from email.message import EmailMessage as _EmailMessage

from .address import AddressList
from .address import prepare_address
from .attachment import Attachment


class Message:
    """A representation of the typical data found in an email message. Can be
    converted to and from an :class:`email.message.EmailMessage`.

    While unlikely, MIME messages can be constructed pretty much arbitrarily.
    This class only represents a "common"/"standard" message structure with
    text, HTML, download attachments, and inline attachments.

    Addresses passed to the various arguments/attributes can be strings in the
    form ``user@domain`` or ``Name <user@domain>``, or instances of
    :class:`email.headerregistry.Address`. If the ``user`` part has non-ASCII
    characters, SMTP servers must support the ``SMTPUTF8`` extension. The domain
    part will be encoded using IDNA.

    :param subject: The text in the subject line of the message.
    :param text: The plain text content.
    :param from_addr: The address to show the message was sent from.
    :param reply_to: Instruct users to reply to this address rather than
        ``from_addr``.
    :param to: The primary recipients, visible and replied to by all other
        recipients.
    :param cc: The secondary recipients, visible and replied to by all other
        recipients.
    :param bcc: Hidden recipients. Not visible or replied to by any other
        recipients. When sending mass email, use this instead of ``to`` and
        ``cc`` to avoid "reply all email storms".
    :param attachments: Downloadable files separate from the content.
    :param html: The HTML text content. ``text`` content should also be
        provided, but if it's not then text is extracted from the HTML.
    :param inline_attachments: Files that can be linked and displayed inside
        HTML content. Only relevant when ``html`` content is provided.
    """

    def __init__(
        self,
        *,
        subject: str | None = None,
        text: str | None = None,
        from_addr: str | Address | None = None,
        reply_to: str | Address | None = None,
        to: list[str | Address] | None = None,
        cc: list[str | Address] | None = None,
        bcc: list[str | Address] | None = None,
        attachments: list[Attachment] | None = None,
        html: str | None = None,
        inline_attachments: list[Attachment] | None = None,
    ):
        self.subject: str | None = subject
        """The text in the subject line of the message."""

        self.text: str | None = text
        """The plain text content."""

        self.html: str | None = html
        """The HTML text content. :attr:`text` content should also be provided,
        but if it's not then text is extracted from the HTML.
        """

        if from_addr:
            self._from_addr: Address | None = prepare_address(from_addr)
        else:
            self._from_addr = None

        if reply_to:
            self._reply_to: Address | None = prepare_address(reply_to)
        else:
            self._reply_to = None

        self._to: AddressList = AddressList(to or ())
        self._cc: AddressList = AddressList(cc or ())
        self._bcc: AddressList = AddressList(bcc or ())

        self.attachments: list[Attachment] = attachments or []
        """Downloadable files separate from the content."""

        self.inline_attachments: list[Attachment] = inline_attachments or []
        """Files that can be linked and displayed inside HTML content. Only
        relevant when :attr:`html` content is provided.
        """

    @property
    def from_addr(self) -> Address | None:
        """The address to show the message was sent from."""
        return self._from_addr

    @from_addr.setter
    def from_addr(self, value: str | Address | None) -> None:
        if value:
            self._from_addr = prepare_address(value)
        else:
            self._from_addr = None

    @property
    def reply_to(self) -> Address | None:
        """Instruct users to reply to this address rather than :attr:`from_addr`."""
        return self._reply_to

    @reply_to.setter
    def reply_to(self, value: str | Address | None) -> None:
        if value:
            self._reply_to = prepare_address(value)
        else:
            self._reply_to = None

    @property
    def to(self) -> AddressList:
        """The primary recipients, visible and replied to by all other recipients."""
        return self._to

    @to.setter
    def to(self, value: cabc.Iterable[str | Address]) -> None:
        self._to[:] = value

    @property
    def cc(self) -> AddressList:
        """The secondary recipients, visible and replied to by all other recipients."""
        return self._cc

    @cc.setter
    def cc(self, value: cabc.Iterable[str | Address]) -> None:
        self._cc[:] = value

    @property
    def bcc(self) -> AddressList:
        """Hidden recipients. Not visible or replied to by any other recipients
        When sending mass email, use this instead of :attr:`to` and :attr:`cc`
        to avoid "reply all email storms".
        """
        return self._bcc

    @bcc.setter
    def bcc(self, value: cabc.Iterable[str | Address]) -> None:
        self._bcc[:] = value

    def to_mime(self) -> _EmailMessage:
        """Convert this :class:`email_simplified.Message` to an
        :class:`email.message.EmailMessage`.
        """
        message = _EmailMessage()

        if self.subject:
            message["Subject"] = self.subject

        if self.from_addr:
            message["From"] = self.from_addr

        if self.reply_to:
            message["Reply-To"] = self.reply_to

        if self.to:
            message["To"] = self.to

        if self.cc:
            message["CC"] = self.cc

        if self.bcc:
            message["BCC"] = self.bcc

        if self.text:
            message.set_content(self.text)
        elif self.html:
            text = _HTMLToText.process(self.html)
            message.set_content(text)

        if self.html:
            message.add_alternative(self.html, subtype="html")
            part = t.cast(_EmailMessage, message.get_payload(1))

            for attachment in self.inline_attachments:
                attachment.add_to_mime(part, inline=True)

        for attachment in self.attachments:
            attachment.add_to_mime(message)

        return message

    @classmethod
    def from_mime(cls, message: _EmailMessage) -> t.Self:
        """Convert an :class:`email.message.EmailMessage` message to a
        :class:`email_simplified.Message`.

        MIME messages can potentially have arbitrary structures. This only
        supports converting from a "standard" structure where a message
        has text, html, inline attachment, and download attachment parts. Can
        be one of the following:

        -   One ``text/plain`` part.
        -   One ``multipart/alternative`` part containing one ``text/plain``
            part then either:

            -   One ``text/html`` part.
            -   One ``multipart/related`` part containing one ``text/html`` part
                then one or more inline attachment parts.

        -   One ``multipart/mixed`` part containing one of the above parts then
            one or more download attachment parts.

        :param message: The MIME part message to convert.
        """
        original = message
        content_type = original.get_content_type()
        parts: list[_EmailMessage]
        text: str
        html: str | None
        attachments: list[Attachment] = []
        inline_attachments: list[Attachment] = []

        if content_type == "multipart/mixed":
            message, *parts = t.cast(list[_EmailMessage], original.get_payload())
            content_type = message.get_content_type()

            for part in parts:
                attachments.append(
                    Attachment(
                        data=part.get_content(),
                        filename=part.get_filename(),
                        mimetype=part.get_content_type(),
                    )
                )

        if content_type == "multipart/alternative":
            text_part: _EmailMessage
            html_part: _EmailMessage
            text_part, html_part = t.cast(list[_EmailMessage], message.get_payload())
            text = text_part.get_content()

            if html_part.get_content_type() == "multipart/related":
                html_part, *parts = t.cast(list[_EmailMessage], html_part.get_payload())

                for part in parts:
                    attachment = Attachment(
                        data=part.get_content(),
                        filename=part.get_filename(),
                        mimetype=part.get_content_type(),
                    )
                    attachment.cid = part["content-id"]
                    inline_attachments.append(attachment)

            html = html_part.get_content()
        else:
            text = message.get_content()
            html = None

        from_addr: AddressHeader = original["from"]
        reply_to: AddressHeader = original["reply-to"]
        to: AddressHeader = original["to"]
        cc: AddressHeader = original["cc"]
        bcc: AddressHeader = original["bcc"]
        return cls(
            subject=original["subject"],
            text=text,
            from_addr=from_addr.addresses[0] if from_addr else None,
            reply_to=reply_to.addresses[0] if reply_to else None,
            to=list(to.addresses) if to else [],
            cc=list(cc.addresses) if cc else [],
            bcc=list(bcc.addresses) if bcc else [],
            attachments=attachments,
            html=html,
            inline_attachments=inline_attachments,
        )


class _HTMLToText(html.parser.HTMLParser):
    """Extract all text data from an HTML document. Used to create text content
    for an email if only HTML content is given.
    """

    def __init__(self) -> None:
        super().__init__(convert_charrefs=True)
        self.data: list[str] = []
        """Accumulates the text parts extracted during parsing."""

    @classmethod
    def process(cls, html: str) -> str:
        """Convert an HTML document into plain text.

        :param html: The HTML to convert.
        """
        parser = cls()
        parser.feed(html)
        return parser.text

    @property
    def text(self) -> str:
        """Join the text parts into a single string."""
        return "".join(self.data)

    def handle_data(self, data: str) -> None:
        self.data.append(data)



from __future__ import annotations

import typing as t
from email.message import EmailMessage as _EmailMessage

from ..message import Message
from .base import EmailHandler


class TestEmailHandler(EmailHandler):
    """Email handler that appends messages to a list rather than sending them.
    Useful when testing.
    """

    __test__ = False  # don't get collected by pytest

    def __init__(self) -> None:
        self.outbox: list[Message | _EmailMessage] = []
        """List of messages that have been sent with this handler."""

    @classmethod
    def from_config(cls, config: dict[str, t.Any]) -> t.Self:
        return cls()

    def send(self, messages: list[Message | _EmailMessage]) -> None:
        self.outbox.extend(messages)

    async def send_async(self, messages: list[Message | _EmailMessage]) -> None:
        self.send(messages)



from __future__ import annotations

import ssl
import typing as t
from contextlib import contextmanager
from email.message import EmailMessage as _EmailMessage
from itertools import islice
from smtplib import SMTP
from smtplib import SMTP_SSL
from smtplib import SMTP_SSL_PORT
from ssl import SSLContext

from ..attachment import local_hostname
from ..message import Message
from .base import EmailHandler


class SMTPEmailHandler(EmailHandler):
    """Email handler that sends with SMTP using Python's built-in
    :mod:`smtplib`. All arguments are optional, :class:`smtplib.SMTP` uses
    default values if something is not given.

    :param host: Host to connect to.
    :param port: Port to connect to.
    :param use_tls: The connection should be established with TLS. Default is
        ``True`` if ``port`` is 465.
    :param use_starttls: The connection should be upgraded to TLS after
        establishing a plain connection first. You should prefer ``use_tls``
        instead if your SMTP server supports it.
    :param tls_context: An :class:`ssl.SSLContext` to use when ``use_tls`` or
        ``use_starttls`` is enabled. Defaults to :func:`ssl.create_default_context`,
        which enables verifying any server cert trusted by the OS. You should
        only need to change this if the server uses a custom cert, or needs a
        client cert sent as well.
    :param timeout: Connection timeout. Default is no timeout.
    :param username: Username to log in with.
    :param password: Password to log in with.
    :param default_from: Default address to send from if the ``Sender`` or
        ``From`` header is not set on a message.
    :param recipients_per_message: If a message has more than this number of
        recipients, split the send calls to batches of this size. This is to
        support servers that have a limit configured. By default, no batching
        is done.
    """

    def __init__(
        self,
        *,
        host: str | None = None,
        port: int | None = None,
        use_tls: bool | None = None,
        use_starttls: bool = False,
        tls_context: SSLContext | None = None,
        timeout: float | None = None,
        username: str | None = None,
        password: str | None = None,
        default_from: str | None = None,
        recipients_per_message: int | None = None,
    ):
        self.host = host
        """Host to connect to."""

        self.port = port
        """Port to connect to."""

        self.timeout = timeout
        """Connection timeout."""

        self.username = username
        """Username to log in with."""

        self.password = password
        """Password to log in with."""

        self.default_from = default_from
        """Default address to send from if the ``Sender`` or ``From`` header is
        not set on a message.
        """

        self.recipients_per_message = recipients_per_message
        """If a message has more than this number of recipients, split the
        low-level send calls to batches of this size. This is to support servers
        that have a limit configured. By default no batching is done.
        """

        if use_tls is None:
            use_tls = port == SMTP_SSL_PORT

        self.use_tls = use_tls
        """The connection should be established with TLS."""

        self.use_starttls: bool = not use_tls and use_starttls
        """The connection should be upgraded to TLS after establishing a plain
        connection first.
        """

        if use_tls or use_starttls and tls_context is None:
            tls_context = ssl.create_default_context()

        self.tls_context = tls_context
        """An :class:`ssl.SSLContext` to use when ``use_tls`` or
        ``use_starttls`` is enabled.
        """

    @classmethod
    def from_config(cls, config: dict[str, t.Any]) -> t.Self:
        """Create a handler from a config dict. Config keys match the
        arguments to :class:`.SMTPEmailHandler`, and are all optional.
        """
        return cls(
            host=config.get("host"),
            port=config.get("port"),
            use_tls=config.get("use_tls"),
            use_starttls=config.get("use_starttls", False),
            tls_context=config.get("tls_context"),
            timeout=config.get("timeout"),
            username=config.get("username"),
            password=config.get("password"),
            default_from=config.get("default_from"),
        )

    @contextmanager
    def connect(self) -> t.Iterator[SMTP]:
        """Context manager that creates an :class:`smtplib.SMTP` client, connects,
        logs in, then closes when exiting the block.
        """
        smtp_cls: type[SMTP | SMTP_SSL] = SMTP
        smtp_args: dict[str, t.Any] = {
            "host": self.host,
            "port": self.port,
            "local_hostname": local_hostname(),
            "timeout": self.timeout,
        }

        if self.use_tls:
            smtp_cls = SMTP_SSL
            smtp_args["context"] = self.tls_context

        with smtp_cls(**smtp_args) as client:
            if self.use_starttls:
                client.starttls(context=self.tls_context)

            if self.username is not None and self.password is not None:
                client.login(self.username, self.password)

            yield client

    def _send_batched(
        self, client: SMTP, message: _EmailMessage, from_addr: str | None
    ) -> None:
        assert self.recipients_per_message is not None
        recipient_fields = (message["to"], message["cc"], message["bcc"])
        recipients = (
            a.addr_spec for f in recipient_fields if f is not None for a in f.addresses
        )

        # itertools.batched for Python < 3.12
        while batch := tuple(islice(recipients, self.recipients_per_message)):
            client.send_message(message, from_addr=from_addr, to_addrs=batch)

    def send(self, messages: list[Message | _EmailMessage]) -> None:
        if not messages:
            return

        with self.connect() as client:
            for message in messages:
                if isinstance(message, Message):
                    mime_message = message.to_mime()
                else:
                    mime_message = message

                from_addr = mime_message["sender"] or mime_message["from"]

                if from_addr:
                    from_addr = from_addr.addresses[0].addr_spec
                else:
                    from_addr = self.default_from

                if self.recipients_per_message:
                    self._send_batched(client, mime_message, from_addr)
                else:
                    client.send_message(mime_message, from_addr=from_addr)



from __future__ import annotations

import asyncio
import importlib.metadata
import pkgutil
import typing as t
from email.message import EmailMessage as _EmailMessage
from inspect import isclass

from ..message import Message


class EmailHandler:
    """Interface for sending email messages. Subclasses will define how to send
    using a service, such as SMTP, an email provider's API, etc.
    """

    def send(self, messages: list[Message | _EmailMessage]) -> None:
        """Send one or more email messages.

        Messages should typically be :class:`.Message`. However, they may also
        be :class:`email.message.EmailMessage` for cases where a non-standard
        MIME construction is needed. It is up to the handler to decide whether
        that is supported based on what their service is capable of.

        :param messages: A list of messages to send.
        """
        raise NotImplementedError

    async def send_async(self, messages: list[Message | _EmailMessage]) -> None:
        """Send one or more email messages, as with :meth:`send`, but in an
        ``async`` context.
        """
        await asyncio.to_thread(self.send, messages)

    @classmethod
    def from_config(cls, config: dict[str, t.Any]) -> t.Self:
        """Create an instance of this handler using arguments from ``config`` as
        needed. This is useful for frameworks when paired with
        :func:`get_handler_class`, allowing the deployment to configure an
        arbitrary handler and its arguments.
        """
        raise NotImplementedError


_handler_classes: dict[str, type[EmailHandler]] = {}


def get_handler_class(name: str | type[EmailHandler]) -> type[EmailHandler]:
    """Get a :class:`.EmailHandler` implementation by name.

    First looks up if the name is in the ``email_simplified.handler`` entry
    point namespace, and loads the referenced class if it is. Libraries
    providing handler implementations should register a handler's name so that
    it's automatically available when installed.

    If an entry point is not found, attempts to treat the name as an import
    path in the form ``module.submodule:handler_class``. This is useful for
    users writing their own handler locally where adding an entry point may not
    be possible. May also be passed an actual handler class, in which case it's
    returned directly.

    :param name: The registered entry point name, import path, or handler class
        to load and return.
    """
    if not isinstance(name, str):
        return name

    if not _handler_classes:
        for ep in importlib.metadata.entry_points(group="email_simplified.handler"):
            obj = ep.load()
            _handler_classes[ep.name] = obj

    if name in _handler_classes:
        return _handler_classes[name]

    try:
        obj = pkgutil.resolve_name(name)
    except ImportError:
        obj = None

    if isclass(obj) and issubclass(obj, EmailHandler):
        _handler_classes[name] = obj
        return obj

    raise ValueError(f"Could not find installed entry point or import: '{name}'.")



from __future__ import annotations

import pytest

from email_simplified import get_handler_class
from email_simplified import SMTPEmailHandler
from email_simplified import TestEmailHandler
from email_simplified.handlers import EmailHandler


@pytest.mark.parametrize(
    ("name", "expect"), [("smtp", SMTPEmailHandler), ("test", TestEmailHandler)]
)
def test_get_entry_point(name: str, expect: type[EmailHandler]) -> None:
    cls = get_handler_class(name)
    assert cls is expect


def test_get_import() -> None:
    cls = get_handler_class("email_simplified.handlers.smtp:SMTPEmailHandler")
    assert cls is SMTPEmailHandler


def test_get_class() -> None:
    cls = get_handler_class(SMTPEmailHandler)
    assert cls is SMTPEmailHandler


def test_get_error() -> None:
    with pytest.raises(ValueError):
        get_handler_class("nothing")


def test_from_config() -> None:
    handler = SMTPEmailHandler.from_config({"port": 1025})
    assert handler.port == 1025



from __future__ import annotations

import asyncio

from email_simplified import Message
from email_simplified import TestEmailHandler


def test_handler() -> None:
    handler = TestEmailHandler()
    handler.send([Message(subject="a")])
    asyncio.run(handler.send_async([Message(subject="b")]))
    assert len(handler.outbox) == 2


def test_config() -> None:
    TestEmailHandler.from_config({"invalid": True})



from __future__ import annotations

from email.headerregistry import Address

import pytest

from email_simplified import Attachment
from email_simplified import Message


def test_init() -> None:
    m = Message(
        subject="a",
        text="a",
        from_addr="a@a.test",
        reply_to="b@a.test",
        to=["c@a.test"],
        cc=["d@a.test"],
        bcc=["e@a.test"],
        html="<p>a</p>",
    )
    assert m.subject == "a"
    assert m.text == "a"
    assert m.html == "<p>a</p>"
    assert m.from_addr == Address(addr_spec="a@a.test")
    assert m.reply_to == Address(addr_spec="b@a.test")
    assert m.to == [Address(addr_spec="c@a.test")]
    assert m.cc == [Address(addr_spec="d@a.test")]
    assert m.bcc == [Address(addr_spec="e@a.test")]


def test_init_empty() -> None:
    m = Message()
    assert m.subject is None
    assert m.text is None
    assert m.html is None
    assert m.from_addr is None
    assert m.reply_to is None
    assert m.to == []
    assert m.cc == []
    assert m.bcc == []


@pytest.mark.parametrize("attr", ["from_addr", "reply_to"])
def test_set_from(attr: str) -> None:
    m = Message()
    setattr(m, attr, "a@a.test")
    assert getattr(m, attr) == Address(addr_spec="a@a.test")
    setattr(m, attr, None)
    assert getattr(m, attr) is None


@pytest.mark.parametrize("attr", ["to", "cc", "bcc"])
def test_set_to(attr: str) -> None:
    m = Message()
    addr = Address(addr_spec="a@a.test")
    setattr(m, attr, [addr])
    assert getattr(m, attr) == [addr]
    setattr(m, attr, [])
    assert getattr(m, attr) == []


def test_mime_headers() -> None:
    m = Message.from_mime(
        Message(
            subject="a",
            text="a",
            from_addr="あ@あ.test",
            reply_to="B <b@a.test>",
            to=["c1@a.test", "c2@a.test"],
            cc=["d1@a.test", "d2@a.test"],
            bcc=["e1@a.test", "e2@a.test"],
        ).to_mime()
    )
    assert m.subject == "a"
    assert str(m.from_addr) == "あ@xn--l8j.test"
    assert str(m.reply_to) == "B <b@a.test>"
    assert len(m.to) == 2
    assert len(m.cc) == 2
    assert len(m.bcc) == 2


def test_to_mime_headers_empty() -> None:
    m = Message.from_mime(Message(text="a").to_mime())
    assert m.subject is None
    assert m.from_addr is None
    assert m.reply_to is None
    assert not m.to
    assert not m.cc
    assert not m.bcc


def test_mime_text() -> None:
    m = Message.from_mime(Message(text="a").to_mime())
    assert m.text == "a\n"
    assert m.html is None


def test_mime_html() -> None:
    m = Message.from_mime(Message(html="<p>b</p>\n<p>c</p>").to_mime())
    assert m.text == "b\nc\n"
    assert m.html == "<p>b</p>\n<p>c</p>\n"


def test_mime_text_html() -> None:
    m = Message.from_mime(Message(text="a", html="<p>b</p>").to_mime())
    assert m.text == "a\n"
    assert m.html == "<p>b</p>\n"


def test_attachments() -> None:
    m = Message.from_mime(
        Message(
            html="a",
            attachments=[Attachment(data="null", filename="a.json")],
            inline_attachments=[Attachment(data=b"null", filename="b.json")],
        ).to_mime()
    )
    assert len(m.attachments) == 1
    assert len(m.inline_attachments) == 1


def test_no_html_attachments() -> None:
    m = Message.from_mime(
        Message(
            text="a",
            attachments=[Attachment(data="a")],
            inline_attachments=[Attachment(data="b")],
        ).to_mime()
    )
    assert len(m.attachments) == 1
    assert len(m.inline_attachments) == 0



from .base import EmailHandler
from .base import get_handler_class
from .smtp import SMTPEmailHandler
from .test import TestEmailHandler

__all__ = [
    "get_handler_class",
    "EmailHandler",
    "SMTPEmailHandler",
    "TestEmailHandler",
]



from __future__ import annotations

import pytest

from email_simplified import Attachment


@pytest.mark.parametrize(
    ("value", "expect"), [("a", "text/plain"), (b"a", "application/octet-stream")]
)
def test_default_mimetype(value: str | bytes, expect: str) -> None:
    assert Attachment(value).mimetype == expect


@pytest.mark.parametrize(
    ("value", "expect"), [("a.html", "text/html"), ("a.invalid", "text/plain")]
)
def test_guess_mimetype(value: str, expect: str) -> None:
    assert Attachment("a", filename=value).mimetype == expect


def test_given_mimetype() -> None:
    assert Attachment("a", mimetype="text/html").mimetype == "text/html"


def test_lazy_cid() -> None:
    data = Attachment("a")
    assert data._cid is None  # pyright: ignore


def test_generate_cid() -> None:
    data = Attachment("a")
    assert data.cid
    assert data._cid is not None  # pyright: ignore
    assert data.cid == data.cid


def test_set_cid() -> None:
    data = Attachment("a")
    data.cid = "test"
    assert data._cid is not None  # pyright: ignore
    assert data.cid == "test"



from __future__ import annotations

import collections.abc as cabc
import email.utils
import sys
import typing as t
from email.headerregistry import Address


class AddressList(list[Address]):
    """A :class:`list` subclass that contains class:`email.headerregistry.Address`
    items, but accepts strings as well for various operations. Domains are IDNA
    encoded if needed.
    """

    def __init__(self, value: cabc.Iterable[str | Address] = (), /) -> None:
        super().__init__(prepare_address(v) for v in value)

    def append(self, value: str | Address, /) -> None:
        super().append(prepare_address(value))

    def extend(self, value: cabc.Iterable[str | Address], /) -> None:
        super().extend(prepare_address(v) for v in value)

    def index(
        self,
        value: str | Address,
        start: t.SupportsIndex = 0,
        stop: t.SupportsIndex = sys.maxsize,
        /,
    ) -> int:
        return super().index(prepare_address(value))

    def count(self, value: str | Address, /) -> int:
        return super().count(prepare_address(value))

    def insert(self, index: t.SupportsIndex, value: str | Address, /) -> None:
        super().insert(index, prepare_address(value))

    def remove(self, value: str | Address, /) -> None:
        super().remove(prepare_address(value))

    @t.overload
    def __setitem__(self, key: t.SupportsIndex, value: str | Address) -> None: ...

    @t.overload
    def __setitem__(self, key: slice, value: cabc.Iterable[str | Address]) -> None: ...

    def __setitem__(
        self,
        key: t.SupportsIndex | slice,
        value: str | Address | cabc.Iterable[str | Address],
    ) -> None:
        if not isinstance(key, slice):
            super().__setitem__(key, prepare_address(value))  # type: ignore[arg-type]
        else:
            super().__setitem__(key, (prepare_address(v) for v in value))  # type: ignore[union-attr]

    def __iadd__(self, other: cabc.Iterable[str | Address]) -> t.Self:  # type: ignore[override, misc]
        return super().__iadd__(prepare_address(v) for v in other)

    def __contains__(self, value: object) -> bool:
        if not isinstance(value, str | Address):
            return False

        return super().__contains__(prepare_address(value))


def prepare_address(address: str | Address) -> Address:
    """Convert a string or :class:`email.headerregistry.Address` to an
    ``Address`` with the domain part IDNA encoded if needed.

    :param address: Address to process.
    """
    if isinstance(address, Address):
        domain = _idna_if_needed(address.domain)

        if domain == address.domain:
            return address

        return Address(address.display_name, address.username, domain)

    name, addr = email.utils.parseaddr(address)
    username, _, domain = addr.rpartition("@")
    domain = _idna_if_needed(domain)
    return Address(name, username, domain)


def _idna_if_needed(domain: str) -> str:
    """If the domain is non-ASCII, IDNA encode it.

    :param domain: The domain to encode.
    """
    try:
        domain.encode("ascii")
    except UnicodeEncodeError:
        return domain.encode("idna").decode("ascii")

    return domain



from __future__ import annotations

from email.headerregistry import Address

import pytest

from email_simplified.address import AddressList
from email_simplified.address import prepare_address


@pytest.mark.parametrize(
    ("value", "expect"),
    [
        ("a@a.test", Address(addr_spec="a@a.test")),
        ("あ@あ.test", Address(username="あ", domain="xn--l8j.test")),
        ("A <a@a.test>", Address(display_name="A", addr_spec="a@a.test")),
        (Address(addr_spec="a@a.test"), Address(addr_spec="a@a.test")),
        (
            Address(username="あ", domain="あ.test"),
            Address(username="あ", domain="xn--l8j.test"),
        ),
    ],
)
def test_prepare_address(value: str | Address, expect: Address) -> None:
    assert prepare_address(value) == expect


def test_list_init() -> None:
    data = AddressList(["a@a.test"])
    assert data[0] == Address(addr_spec="a@a.test")


def test_list_init_empty() -> None:
    data = AddressList()
    assert not data


def test_list_append() -> None:
    data = AddressList()
    data.append("a@a.test")
    assert data[0] == Address(addr_spec="a@a.test")


def test_list_extend() -> None:
    data = AddressList()
    data.extend(["a@a.test", "b@a.test"])
    assert len(data) == 2
    assert all(isinstance(x, Address) for x in data)


def test_list_index() -> None:
    data = AddressList(["b@a.test", "a@a.test"])
    assert data.index("a@a.test") == 1


def test_list_count() -> None:
    data = AddressList(["a@a.test", "a@a.test"])
    assert data.count("a@a.test") == 2


def test_list_insert() -> None:
    data = AddressList(["a@a.test"])
    data.insert(0, "b@a.test")
    assert data == [Address(addr_spec="b@a.test"), Address(addr_spec="a@a.test")]


def test_list_remove() -> None:
    data = AddressList(["b@a.test", "a@a.test"])
    data.remove("b@a.test")
    assert data == [Address(addr_spec="a@a.test")]


def test_list_setitem() -> None:
    data = AddressList(["b@a.test", "a@a.test"])
    data[1] = "c@a.test"
    assert data[1] == Address(addr_spec="c@a.test")
    data[:] = ["d@a.test", "e@a.test"]
    assert data == [Address(addr_spec="d@a.test"), Address(addr_spec="e@a.test")]


def test_list_iadd() -> None:
    data = AddressList(["a@a.test"])
    data += ["b@a.test", "c@a.test"]
    assert len(data) == 3


def test_list_contains() -> None:
    data = AddressList(["a@a.test"])
    assert "a@a.test" in data
    assert "b@a.test" not in data
    assert object() not in data



from .attachment import Attachment
from .handlers.base import get_handler_class
from .handlers.smtp import SMTPEmailHandler
from .handlers.test import TestEmailHandler
from .message import Message

__all__ = [
    "get_handler_class",
    "Attachment",
    "Message",
    "SMTPEmailHandler",
    "TestEmailHandler",
]
